package kd.hrmp.hbjm.business.task;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import kd.bos.context.RequestContext;
import kd.bos.dataentity.entity.DynamicObject;
import kd.bos.dataentity.entity.DynamicObjectCollection;
import kd.bos.dataentity.metadata.dynamicobject.DynamicObjectType;
import kd.bos.dataentity.serialization.SerializationUtils;
import kd.bos.exception.KDBizException;
import kd.bos.exception.KDException;
import kd.bos.logging.Log;
import kd.bos.logging.LogFactory;
import kd.bos.orm.query.QCP;
import kd.bos.orm.query.QFilter;
import kd.bos.orm.util.CollectionUtils;
import kd.bos.schedule.executor.AbstractTask;
import kd.hr.hbp.business.servicehelper.HRMServiceHelper;
import kd.hr.hbp.business.servicehelper.HRProducerServiceHelper;
import kd.hr.hbp.common.constants.HRBaseConstants;
import kd.hr.hbp.common.constants.HRHisBaseConstants;
import kd.hrmp.hbjm.business.domain.repository.HBJMMsgRepository;
import kd.hrmp.hbjm.common.constants.ChangeEventConstants;
import kd.hrmp.hbjm.common.constants.PublishConstants;
import kd.hrmp.hbjm.common.util.HBJMUtil;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class PublishTask extends AbstractTask implements Runnable {
    private static final Log logger = LogFactory.getLog(PublishTask.class);

    private List<DynamicObject> msgList;
    private Map<Long, List<Map<String, String>>> actionConsumerMap = Maps.newHashMapWithExpectedSize(HRBaseConstants.INT_EIGHT);

    public PublishTask() {}

    public PublishTask(List<DynamicObject> msgList) {
        this.msgList = msgList;
    }

    @Override
    public void run() {
        publish(msgList, false);
    }

    /**
     * 同步订阅者的消费状态
     *
     * @param requestContext
     * @param map
     * @throws KDException
     */
    @Override
    public void execute(RequestContext requestContext, Map<String, Object> map) throws KDException {
        // 查询出待重同步的
        DynamicObject[] msgList = HBJMMsgRepository.getInstance().querySyncData();
        syncRecStatus(msgList);

        // 查询出推送状态为待推送的
        DynamicObject[] sendstateSyncList = HBJMMsgRepository.getInstance().querySendstateSyncData();
        syncSendstate(sendstateSyncList);
    }

    private void syncRecStatus(DynamicObject[] msgList) {
		if(msgList == null){
            return;
        }
        Map<String, Map<String, Object>> msgNoToDataMapMap = getMsgCenterData(msgList);
        for (DynamicObject msg : msgList) {
            String msgNo = msg.getString("number");

            // 分录为空，表示没有订阅者
            DynamicObjectCollection entry = msg.getDynamicObjectCollection("entryentity");
            if (CollectionUtils.isEmpty(entry)) {
                continue;
            }

            Map<String, Object> dataMap = msgNoToDataMapMap.get(msgNo);
            if (dataMap == null) {
                continue;
            }

            List<Map<String, Object>> subList = (List<Map<String, Object>>)dataMap.get("dataList");
            for (Map<String, Object> sub : subList) {
                long subscribeBD = (Long)sub.get("subscribeBD");
                String recStatus = (String)sub.get("recStatus");
                for (DynamicObject entryDyn : entry) {
                    long entrySubscribebd = entryDyn.getLong("subscribebd.id");
                    if (entrySubscribebd == subscribeBD) {
                        entryDyn.set("recstatus", recStatus);
                    }
                }
            }
        }
        HBJMMsgRepository.getInstance().saveBatch(Arrays.asList(msgList));
    }

    private Map<String, Map<String, Object>> getMsgCenterData(DynamicObject[] msgList){
        if (msgList == null || msgList.length == 0) {
            return new HashMap<>();
        }

        // 这一段要等到中台提供批量查询接口后，再进行优化
        List<Map<String, Object>> paramsList = Lists.newArrayListWithExpectedSize(msgList.length);
        for (DynamicObject msg : msgList) {
            String msgNo = msg.getString("number");
            long actionId = getMsgActionId(msg);

            Map<String, Object> paramMap = new HashMap<>();
            paramMap.put("msgNumber", msgNo);
            paramMap.put("actionId", actionId);
            paramsList.add(paramMap);
        }
        if (paramsList.size() == 0) {
            return new HashMap<>();
        }
        List<Map<String, Object>> dataList = getSubRecStatus(paramsList);
        return dataList.stream().collect(
                Collectors.toMap(o -> o.get("msgNumber").toString(), o -> o, (o1, o2) -> o2));
    }

    private void syncSendstate(DynamicObject[] msgList) {
        Map<String, Map<String, Object>> msgNoToDataMapMap = getMsgCenterData(msgList);
        for (DynamicObject msg : msgList) {
            String msgNo = msg.getString("number");

            Map<String, Object> dataMap = msgNoToDataMapMap.get(msgNo);
            if (dataMap != null) {
                msg.set("sendstate", PublishConstants.PUBLISH_SUCCESS);
            }
        }
        HBJMMsgRepository.getInstance().saveBatch(Arrays.asList(msgList));
    }

    public void publish(List<DynamicObject> msgList, boolean retry) {
        if (CollectionUtils.isEmpty(msgList)) {
            return;
        }
        Date now = new Date();
        List<DynamicObject> saveList = new ArrayList<>(msgList.size());
        for (DynamicObject msg : msgList) {
            if (retry) {
                msg.set("retrytimes", msg.getInt("retrytimes") + 1);
            }
            int step = 0;
            long actionId = getMsgActionId(msg);
            try {
                Map<String, Object> actionInfo = getPublishInfoMap(msg, now);
                // 先getSubinfo，再发消息，再setConsumerEntry 。
                // 防止消息发送成功，但是getSubinfo异常，导致重发时消息已存在。
                // 订阅者信息只在发送成功消息后，保存到分录去，发失败时不保存。
                List<Map<String, String>> subInfoList = getSubinfo(actionId);
                step++;
                HRProducerServiceHelper.publishAction(actionInfo);
                setConsumerEntry(msg, subInfoList);
                msg.set("sendstate", PublishConstants.PUBLISH_SUCCESS);
                // 成功后清空失败原因
            } catch (KDBizException ex) {
                // 已经发送过了
                if (step > 0 && isSendSuccess(msg.getString("number"), actionId)) {
                    msg.set("sendstate", PublishConstants.PUBLISH_SUCCESS);
                } else {
                    String errorMsg = ex.getMessage();
                    msg.set("sendstate", PublishConstants.PUBLISH_FAIL);
                    logger.error(errorMsg);
                    logger.error("pubmsg-error");
                }
            }
            msg.set("lastsendtime", now);
            saveList.add(msg);
        }
        HBJMMsgRepository.getInstance().saveBatch(saveList);
    }

    private List<Map<String, String>> getSubinfo(long actionId) {
        List<Map<String, String>> subInfoList = actionConsumerMap.get(actionId);
        if (subInfoList == null) {
            Map<String, Object> pubAndSubInfo = getSimplePubAndSubInfo(PublishConstants.msgPubNoMap.get(actionId));
            subInfoList = (List<Map<String, String>>)pubAndSubInfo.get("subInfoList");
            if (subInfoList == null) {
                subInfoList = Lists.newArrayListWithExpectedSize(HRBaseConstants.INT_EIGHT);
            }
            actionConsumerMap.put(actionId, subInfoList);
            return subInfoList;
        }
        return subInfoList;
    }

    private void setConsumerEntry(DynamicObject msg, List<Map<String, String>> subInfoList) {
        if (CollectionUtils.isEmpty(subInfoList)) {
            return;
        }
        DynamicObjectCollection consumerEntry = msg.getDynamicObjectCollection("entryentity");
        for (Map<String, String> consumer : subInfoList) {
            DynamicObjectType entryType = consumerEntry.getDynamicObjectType();
            DynamicObject row = new DynamicObject(entryType);
            row.set("consumercloud", consumer.get("consumerCloudId"));
            row.set("consumerapp", consumer.get("consumerAppId"));
            row.set("subscribebd", consumer.get("businessTypeId"));
            row.set("recstatus", PublishConstants.RECSTATUS_SYNC);
            consumerEntry.add(row);
        }
    }

    private Map<String, Object> getSimplePubAndSubInfo(String msgPubNo) {
        return HRMServiceHelper.invokeBizService("hrmp", "hrcs", "IHRCSMsgService", "getSimplePubAndSubInfo", msgPubNo);
    }

    private List<Map<String, Object>> getSubRecStatus(List<Map<String, Object>> paramsList) {
        try {
            return HRMServiceHelper.invokeBizService("hrmp", "hrcs", "IHRCSMsgService", "getMsgCenterListInfo",
                    paramsList);
        } catch (Exception ex) {
            return Lists.newArrayListWithExpectedSize(HRBaseConstants.INT_EIGHT);
        }
    }

    private boolean isSendSuccess(String msgNumber, long actionId) {
        List<QFilter> filters = new ArrayList<>(4);
        filters.add(new QFilter("number", QCP.equals, msgNumber));
        filters.add(new QFilter("action", QCP.equals, actionId));
        return HRMServiceHelper.invokeBizService("hrmp", "hrcs", "IHRCSMsgService", "validateMsgIsDuplicate",
            new Object[] {filters.toArray(new QFilter[0])});
    }

    private Map<String, Object> getPublishInfoMap(DynamicObject msg, Date now) {
        String msgNumber = msg.getString("number");
        long actionId = getMsgActionId(msg);
        String msgTitle = msg.getString("msgtitle");
        String msgTag = msg.getString("msgtag");
        String msgDesc = msg.getString("msgdesc");

        Map<String, Object> actionInfo = Maps.newHashMapWithExpectedSize(
                HRBaseConstants.INT_CAPACITY);
        actionInfo.put("msgNumber", msgNumber);
        actionInfo.put("actionId", actionId);
        actionInfo.put("senderId", 1L);
        actionInfo.put("sendTime", now);
        actionInfo.put("msgPubNo", PublishConstants.msgPubNoMap.get(actionId));
        actionInfo.put("msgTitle", msgTitle);
        actionInfo.put("params", SerializationUtils.serializeToBase64(getParams(msg)));
        actionInfo.put("msgDesc", msgDesc);
        actionInfo.put("msgTag", msgTag);
        return actionInfo;
    }

    private long getMsgActionId(DynamicObject msg) {
        return PublishConstants.operateActionMap.get(HBJMUtil.getBasedataPkId(msg, "changeoperate"));
    }

    private Map<String, Object> getParams(DynamicObject msg) {
        long changeoperate = HBJMUtil.getBasedataPkId(msg, "changeoperate");
        Map<String, Object> params = Maps.newHashMapWithExpectedSize(HRBaseConstants.INT_EIGHT);
        params.put("number", msg.getString("job.number"));
        params.put("name", msg.getString("job.name"));
        params.put("id", msg.getLong("job.id"));
        params.put("changeoperate", changeoperate);
        params.put("changedate", msg.getDate("effetdate"));
        params.put("boid", msg.getDynamicObject("job").getLong("boid"));
        // 消息编码就是用了事务id的值
        params.put("eventId", msg.getString("number"));
        if (changeoperate == ChangeEventConstants.UPDATE) {
            DynamicObjectCollection changesubsceneCol = msg.getDynamicObjectCollection("changesubscene");
            if (changesubsceneCol != null && changesubsceneCol.size() > 0) {
                List<String> subsceneIds = Lists.newArrayListWithExpectedSize(changesubsceneCol.size());
                for (DynamicObject subscene : changesubsceneCol) {
                    subsceneIds.add(subscene.getString("fbasedataid"));
                }
                params.put("changesubscene", String.join(",", subsceneIds));
            }
            // 判断插入的是否是中间版本
            if (!HRHisBaseConstants.DATA_STATUS_EFFECTING
                .equals(msg.getDynamicObject("job").getString(HRHisBaseConstants.FIELD_DATASTATUS))) {
                params.put("ispreversion", Boolean.TRUE);
            }
        } else if (changeoperate == ChangeEventConstants.DISABLE) {
            params.put("enable", msg.getDynamicObject("job").getString("enable"));
        }
        return params;
    }

}
